#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

void print_ratios( int, int );
long file_size( char *name );
int SavePrgmFile(unsigned char *pData, unsigned nLength, const char *szFilename, FILE *out, char *szExt);

int main( argc, argv )
int argc;
char *argv[];
{
	FILE *output;
	FILE *input;
	unsigned char *data;
	unsigned char *szext;
	int nlength, outted;

	setbuf( stdout, NULL );
	if ( argc != 3)
	{
		puts ("\nUsage:  LITE86 infile outfile\n\tIf \"outfile\" has extension 86P or 86S, it is automatically\n\tconverted to TI format. 86P files will be self-extracting.");
		return 1;
	}

	input = fopen( argv[ 1 ], "rb" );
	if ( input == NULL )
	{
		puts ("Error opening file for input");
		return 2;
	}
	output = fopen( argv[ 2 ] , "wb" );
	if ( output == NULL )
	{
		puts ("Error opening file for output");
		return 2;
	}
	printf( "\nCompressing %s to %s\n", argv[ 1 ], argv[ 2 ] );
	puts( "Using LITE86 II written by Kirk Meyer" );

    fseek( input, 0, SEEK_END );
	nlength = ftell ( input );
	fseek( input, 0, SEEK_SET );
	if ((data = malloc (nlength)) == NULL)
	{
		puts ("Error allocating memory");
		return 2;
	}
	fread (data, 1, nlength, input);
	fclose( input );

	szext = strchr (argv[2], 0);
	if (*(szext - 1) == '.')
		szext -= 1;
	else if (*(szext - 2) == '.')
		szext -= 2;
	else if (*(szext - 3) == '.')
		szext -= 3;
	else if (*(szext - 4) == '.')
		szext -= 4;
	outted = SavePrgmFile (data, nlength, argv[2], output, szext);
	free (data);

	fclose( output );
	print_ratios( nlength, outted );

	return( 0 );
}

/*
 * This routine is used by main to print out get the size of a file after
 * it has been closed.  It does all the work, and returns a long.  The
 * main program gets the file size for the plain text, and the size of
 * the compressed file, and prints the ratio.
 */
#ifndef SEEK_END
#define SEEK_END 2
#endif

long file_size( name )
char *name;
{
    long eof_ftell;
    FILE *file;

	file = fopen( name, "r" );
    if ( file == NULL )
        return( 0L );
    fseek( file, 0L, SEEK_END );
	eof_ftell = ftell( file );
    fclose( file );
    return( eof_ftell );
}

/*
 * This routine prints out the compression ratios after the input
 * and output files have been closed.
 */
void print_ratios( input_size, output_size )
int input_size;
int output_size;
{
	int ratio;

	if ( input_size == 0 )
		input_size = 1;
	ratio = 100 - (int) ( output_size * 100L / input_size );
	printf( "\nInput bytes:        %ld\n", input_size );
	printf( "Output bytes:       %ld\n", output_size );
	if ( output_size == 0 )
		output_size = 1;
	printf( "Compression ratio:  %d%%\n", ratio );
	if (input_size < output_size)
		printf ("WARNING! Compression was not effective.\n");
}

/************************** End of MAIN-C.C **************************/

#define WINDOW_SIZE 2048
#define MAX_MATCH 34

int SavePrgmFile(unsigned char *pData, unsigned nLength, const char *szFilename, FILE *out, char *szExt)
{
	unsigned checksum, lenx, leno;
	unsigned wdval, total;
	unsigned i, j, k;
	int fsave1, fsave2, fload;
	unsigned maxmatch, maxlen;
	unsigned type = 0, addr = 0xD748;
	unsigned bitmask = 1, progress = 0x80;
	unsigned char pbuff[256], *p;
	unsigned char ctemp;
	unsigned char unfx[] = {
		0x21,0x00,0x00,0x11,0x00,0x90,0x01,0x00,0x00,0xED,0xB0,0x21,0x2E,0x90,0x11,
		0x00,0x00,0xD5,0xC3,0x00,0x90,0x7E,0x23,0x37,0x8F,0x06,0x87,0x28,0xF8,0x38,0x04,
		0xED,0xA0,0x18,0xF7,0x47,0x0E,0x1F,0x7E,0xB9,0xC8,0xC5,0x23,0x46,0x23,0xE5,0x68,
		0x47,0xB1,0x07,0x07,0x07,0x67,0x78,0xA1,0xC6,0x03,0x4F,0x06,0x00,0x19,0xED,0xB0,
		0xE1,0xF1,0x18,0xD7 };
	if (!_stricmp(szExt,".86P"))
		type = 0x12;
	if (!_stricmp(szExt,".86S"))
		type = 0x0C;

	(*szExt) = 0;

	switch (type)
	{
	case 0x12:
		puts ("Creating TI-86 program...");
		break;
	case 0x0C:
		puts ("Creating TI-86 string...");
		break;
	default:
		puts ("Creating binary file...");
	}

	if (type)
	{
		fwrite("**TI86**\x1a\x0a\0Lite86 II compressed with Asmide86 v3.1\0\0\0\0", 1, 55, out);
		// we skipped the total file length, we come back to it later!

		total = checksum = 0;

		lenx = 2;

		wdval = 0x000C;
		fwrite(&wdval, 2, 1, out);	/* always 0C,00 */
		checksum += wdval;
			
		fsave1 = ftell (out);
		fseek (out, 2, SEEK_CUR);

		fwrite(&type, 1, 1, out);	/* program type */
		checksum += type;

		if (!(p = strrchr(szFilename, '\\')))
			strncpy(pbuff, szFilename, 8);
		else
			strncpy(pbuff, p + 1, 8);
		pbuff[8] = '\0';

		wdval = strlen(pbuff);
		fwrite(&wdval, 1, 1, out);
		checksum += wdval;
		fwrite(pbuff, 1, 8, out);
		for (i = 0; i < 8; i++)
			checksum += pbuff[i];

		fsave2 = ftell (out);
		fseek (out, 4, SEEK_CUR);
	}
	if (type == 0x12)
	{
		wdval = 0x288E;		/* binary asm program */
		fwrite(&wdval, 2, 1, out);
		checksum += 0x28 + 0x8E;
		lenx += 2;

		if (nLength >= 4)
		{
			if (!pData[0] && pData[1] == 0xC3)
			{
				addr = (pData[3] << 8) + pData[2];
				fwrite (pData, 1, addr - 0xD748, out);
				for (i = 0xD748; i < addr; i++)
				{
					checksum+=pData[i-0xD748];
				}
				lenx+=(addr-0xD748);
				pData+=(addr-0xD748);
			}
		}

		fseek (out, sizeof (unfx), SEEK_CUR);
	}

	leno = lenx;
	pbuff[0] = 0;
	for (i = 0; i < nLength; i++)
	{
		for (maxlen = 0, j = (i < WINDOW_SIZE ? 0 : i - WINDOW_SIZE); j < i; j++)
		{
			for (k = i; k < (i + MAX_MATCH >= nLength ? nLength : i + MAX_MATCH); k++)
			{
				if (pData[k] != pData[j + k - i])
					break;
			}
			k -= i;
			if (k == MAX_MATCH && (i - j) >= WINDOW_SIZE - 256)
				k--;
			if (k >= maxlen)
			{
				maxlen = k;
				maxmatch = j;
			}
		}					

		if (maxlen < 3)
		{
			progress >>= 1;
			pbuff[bitmask++] = pData[i];
			checksum += pData[i];
			lenx++;
			
			if (!progress)
			{
				checksum += pbuff[0];
				lenx++;
				fwrite (pbuff + 1, 1, bitmask - 1, out);
				pbuff[0] = 0;
				bitmask = 1;
				progress = 0x80;
			}
		}
		else
		{
			pbuff[0] |= progress;
			progress >>= 1;
			wdval = maxmatch - i;
			ctemp = ((wdval & 0x700) >> 3) + (maxlen - 3);
			pbuff[bitmask++] = ctemp;
			checksum += ctemp;
			ctemp = wdval & 255;
			pbuff[bitmask++] = ctemp;
			checksum += ctemp;
			lenx+=2;
			i+=(maxlen-1);

			if (!progress)
			{
				checksum += pbuff[0];
				lenx++;
				fwrite (pbuff + 1, 1, bitmask - 1, out);
				pbuff[0] = 0;
				bitmask = 1;
				progress = 0x80;
			}
		}
	}
	//pbuff[0] |= progress;
	//pbuff[bitmask++] = 0x1F;
	//checksum += (pbuff[0] + 0x1F);
	//lenx+=2;
	fwrite (pbuff + 1, 1, bitmask - 1, out);
	leno = lenx - leno;
	
	if (type == 0x12)
	{
		fload = ftell (out);
		fseek (out, 75 + addr - 0xD748, SEEK_SET);
		unfx[1] = (addr + 21) & 255;
		unfx[2] = (addr + 21) >> 8;
		wdval = leno + sizeof (unfx) - 21;
		unfx[7] = wdval & 255;
		unfx[8] = wdval >> 8;
		unfx[15] = addr & 255;
		unfx[16] = addr >> 8;
		fwrite (unfx, 1, sizeof (unfx), out);
		for (i = 0; i < sizeof (unfx); i++)
		{
			checksum+=unfx[i];
		}
		lenx+=sizeof (unfx);
		fseek (out, fload, SEEK_SET);
	}

	if (type)
	{
		total+=(lenx+16);
		fload = ftell (out);
		fseek (out, fsave1, SEEK_SET);
		fwrite(&lenx, 2, 1, out);
		checksum += (lenx & 0x00FF) + (lenx>>8);
		fseek (out, fsave2, SEEK_SET);
		fwrite(&lenx, 2, 1, out);
		checksum += (lenx & 0x00FF) + (lenx>>8);
		lenx-=2;
		fwrite(&lenx, 2, 1, out);
		checksum += (lenx & 0x00FF) + (lenx>>8);
		fseek (out, fload, SEEK_SET);			
		fwrite(&checksum, 2, 1, out);
		fseek(out, 53, SEEK_SET);
		fwrite(&total, 2, 1, out);
		fclose (out);
	}

	if (type == 0x12)
	{
		return (leno + sizeof (unfx));
	}
	return(leno);
}
